(function (global, factory) {
  typeof exports === 'object' && typeof module !== 'undefined' ? factory(exports) :
  typeof define === 'function' && define.amd ? define(['exports'], factory) :
  (factory((global.Unistore = {})));
}(this, function (exports) { 'use strict';

  var babelHelpers = {};
  babelHelpers.typeof = typeof Symbol === "function" && typeof Symbol.iterator === "symbol" ? function (obj) {
    return typeof obj;
  } : function (obj) {
    return obj && typeof Symbol === "function" && obj.constructor === Symbol ? "symbol" : typeof obj;
  };

  babelHelpers.classCallCheck = function (instance, Constructor) {
    if (!(instance instanceof Constructor)) {
      throw new TypeError("Cannot call a class as a function");
    }
  };

  babelHelpers.createClass = function () {
    function defineProperties(target, props) {
      for (var i = 0; i < props.length; i++) {
        var descriptor = props[i];
        descriptor.enumerable = descriptor.enumerable || false;
        descriptor.configurable = true;
        if ("value" in descriptor) descriptor.writable = true;
        Object.defineProperty(target, descriptor.key, descriptor);
      }
    }

    return function (Constructor, protoProps, staticProps) {
      if (protoProps) defineProperties(Constructor.prototype, protoProps);
      if (staticProps) defineProperties(Constructor, staticProps);
      return Constructor;
    };
  }();

  babelHelpers.inherits = function (subClass, superClass) {
    if (typeof superClass !== "function" && superClass !== null) {
      throw new TypeError("Super expression must either be null or a function, not " + typeof superClass);
    }

    subClass.prototype = Object.create(superClass && superClass.prototype, {
      constructor: {
        value: subClass,
        enumerable: false,
        writable: true,
        configurable: true
      }
    });
    if (superClass) Object.setPrototypeOf ? Object.setPrototypeOf(subClass, superClass) : subClass.__proto__ = superClass;
  };

  babelHelpers.possibleConstructorReturn = function (self, call) {
    if (!self) {
      throw new ReferenceError("this hasn't been initialised - super() hasn't been called");
    }

    return call && (typeof call === "object" || typeof call === "function") ? call : self;
  };

  babelHelpers;

  // https://github.com/sindresorhus/object-assign

  var hasOwnProperty = Object.prototype.hasOwnProperty;

  function toObject(val) {
    if (val === null || val === undefined) {
      throw new TypeError('Object.assign cannot be called with null or undefined');
    }

    return Object(val);
  }

  var objectAssign = Object.assign || function (target, source) {
    var to = toObject(target);
    var from = undefined;

    for (var s = 1, len = arguments.length; s < len; s++) {
      from = Object(arguments[s]);

      for (var key in from) {
        if (hasOwnProperty.call(from, key)) {
          to[key] = from[key];
        }
      }
    }

    return to;
  };

  function clone(object) {
    return objectAssign({}, object);
  }

  var RESTAdapter = function () {
    function RESTAdapter() {
      babelHelpers.classCallCheck(this, RESTAdapter);

      this._endpointMap = {};
      if (this.constructor.name === 'RESTAdapter') throw TypeError('RESTAdapter is a base class');
    }

    babelHelpers.createClass(RESTAdapter, [{
      key: 'registerEndpoints',
      value: function registerEndpoints(type, endpoints) {
        this._endpointMap[type] = endpoints;
        return this;
      }
    }, {
      key: 'isSuccess',
      value: function isSuccess(status) {
        return !!(status && status < 400);
      }
    }, {
      key: 'endpointFor',
      value: function endpointFor(type, params) {
        var typeEndpoints = this._endpointMap[type];

        if (typeEndpoints) {
          var endpointOne = typeEndpoints.one;
          var endpointMany = typeEndpoints.many;
          var typeofParams = typeof params === 'undefined' ? 'undefined' : babelHelpers.typeof(params);
          var dynamicSegments = undefined,
              keySegment = undefined,
              keyName = undefined,
              keyValue = undefined;

          if (typeofParams === 'undefined') {
            return endpointMany;
          }

          if (endpointOne) {
            dynamicSegments = endpointOne.match(/:\w+/g);
            if (dynamicSegments) {
              keySegment = dynamicSegments[dynamicSegments.length - 1];
              keyName = keySegment.slice(1);
              keyValue = params[keyName];
            }
          }

          if (typeofParams === 'object') {
            if (keyValue === undefined) {
              return this.buildUrl(endpointMany, params);
            }

            delete params[keyName];
            return this.buildUrl(endpointOne.replace(keySegment, keyValue), params);
          }

          return endpointOne.replace(keySegment, params);
        }
      }
    }, {
      key: 'buildUrl',
      value: function buildUrl(url, data) {
        var query = [];
        for (var prop in data) {
          if (data.hasOwnProperty(prop)) {
            query.push(encodeURIComponent(prop) + '=' + encodeURIComponent(data[prop]));
          }
        }

        if (query.length) {
          var indexOfQuestionMark = url.indexOf('?');
          if (indexOfQuestionMark === -1) {
            url += '?';
          } else if (indexOfQuestionMark !== url.length - 1) {
            url += '&';
          }

          url += query.join('&');
        }
        return url;
      }
    }]);
    return RESTAdapter;
  }();

  var FetchAdapter = function (_RESTAdapter) {
    babelHelpers.inherits(FetchAdapter, _RESTAdapter);

    function FetchAdapter() {
      babelHelpers.classCallCheck(this, FetchAdapter);
      return babelHelpers.possibleConstructorReturn(this, Object.getPrototypeOf(FetchAdapter).apply(this, arguments));
    }

    babelHelpers.createClass(FetchAdapter, [{
      key: 'get',
      value: function get(type, params) {
        var _this2 = this;

        var url = this.endpointFor(type, params);

        return fetch(url).then(function (response) {
          return _this2.isSuccess(response.status) ? response.json() : Promise.reject(response);
        });
      }
    }]);
    return FetchAdapter;
  }(RESTAdapter);

  var XHRAdapter = function (_RESTAdapter) {
    babelHelpers.inherits(XHRAdapter, _RESTAdapter);

    function XHRAdapter() {
      babelHelpers.classCallCheck(this, XHRAdapter);
      return babelHelpers.possibleConstructorReturn(this, Object.getPrototypeOf(XHRAdapter).apply(this, arguments));
    }

    babelHelpers.createClass(XHRAdapter, [{
      key: 'get',
      value: function get(type, params) {
        var _this2 = this;

        return new Promise(function (resolve, reject) {
          var url = _this2.endpointFor(type, params);
          var xhr = new XMLHttpRequest();

          xhr.onreadystatechange = function () {
            if (xhr.readyState === 4) {
              if (_this2.isSuccess(xhr.status)) {
                try {
                  resolve(JSON.parse(xhr.responseText));
                } catch (e) {
                  reject(e);
                }
              } else {
                reject(xhr);
              }
            }
          };

          xhr.open('GET', url);
          xhr.send();
        });
      }
    }]);
    return XHRAdapter;
  }(RESTAdapter);

  var http = undefined;
  var https = undefined;
  if (typeof window === 'undefined' && typeof require !== 'undefined') {
    http = require('http');
    https = require('https');
  }

  function httpProtocolModuleForUrl(url) {
    return url.match(/^(?:http(s?)\:\/\/)?/i)[0] === 'https://' ? https : http;
  }

  var HTTPAdapter = function (_RESTAdapter) {
    babelHelpers.inherits(HTTPAdapter, _RESTAdapter);

    function HTTPAdapter() {
      babelHelpers.classCallCheck(this, HTTPAdapter);
      return babelHelpers.possibleConstructorReturn(this, Object.getPrototypeOf(HTTPAdapter).apply(this, arguments));
    }

    babelHelpers.createClass(HTTPAdapter, [{
      key: 'get',
      value: function get(type, params) {
        var _this2 = this;

        return new Promise(function (resolve, reject) {
          var url = _this2.endpointFor(type, params);

          httpProtocolModuleForUrl(url).get(url, function (response) {
            var status = response.statusCode;
            var body = '';

            // normalizes 'status' property with other adapters
            response.status = status;

            if (_this2.isSuccess(status)) {
              response.on('data', function (data) {
                body += data;
              });
              response.on('end', function () {
                try {
                  resolve(JSON.parse(body));
                } catch (e) {
                  reject(e);
                }
              });
            } else {
              reject(response);
            }
          }).on('error', reject);
        });
      }
    }]);
    return HTTPAdapter;
  }(RESTAdapter);

  function autoAdapter() {
    var win = typeof window !== 'undefined' && window;

    if (win) {
      if (typeof win.fetch !== 'undefined') {
        return new FetchAdapter();
      }
      return new XHRAdapter();
    }

    return new HTTPAdapter();
  }

  var Store = function () {
    function Store() {
      var opts = arguments.length <= 0 || arguments[0] === undefined ? {} : arguments[0];
      babelHelpers.classCallCheck(this, Store);

      this.primaryKey = opts.primaryKey || 'id';
      this.adapter = opts.adapter || autoAdapter();
      this._registry = {};
      this._data = {};
      this._transforms = [];
    }

    babelHelpers.createClass(Store, [{
      key: 'register',
      value: function register(name, opts) {
        var data = this._data;
        var endpoints = opts && opts.endpoints;

        this._registry[name] = opts || {};
        data[name] = data[name] || {};

        if (endpoints) {
          this.adapter.registerEndpoints(name, endpoints);
        }

        return this;
      }
    }, {
      key: 'lookup',
      value: function lookup(name) {
        return this._registry[name];
      }
    }, {
      key: 'transform',
      value: function transform(fnc) {
        if (typeof fnc === 'function') {
          this._transforms.push(fnc);
        }
      }
    }, {
      key: 'get',
      value: function get(type, params) {
        if (this._isManyRequest(type, params)) {
          return this._getMany.apply(this, arguments);
        }
        return this._getByKey.apply(this, arguments);
      }
    }, {
      key: 'push',
      value: function push(type, records) {
        if (Array.isArray(records)) {
          return this._pushMany.apply(this, arguments);
        }
        return this._pushOne.apply(this, arguments);
      }
    }, {
      key: 'find',
      value: function find(type, params) {
        if (this._isManyRequest(type, params)) {
          return this._findMany.apply(this, arguments);
        }
        return this._findByKey.apply(this, arguments);
      }
    }, {
      key: 'fetch',
      value: function fetch(type, params) {
        if (this._isManyRequest(type, params)) {
          return this._fetchMany.apply(this, arguments);
        }
        return this._fetchByKey.apply(this, arguments);
      }

      /**
       * Private
       */

    }, {
      key: '_isManyRequest',
      value: function _isManyRequest(type, params) {
        var typeofParams = typeof params === 'undefined' ? 'undefined' : babelHelpers.typeof(params);
        return typeofParams === 'undefined' || typeofParams === 'object' && !params[this._primaryKeyForType(type)];
      }
    }, {
      key: '_primaryKeyForType',
      value: function _primaryKeyForType(type) {
        var typeOpts = this.lookup(type);
        return typeOpts && typeOpts.primaryKey || this.primaryKey;
      }
    }, {
      key: '_getByKey',
      value: function _getByKey(type, key) {
        var typeStore = this._data[type];
        var record = typeStore && typeStore[key];
        if (record) {
          return clone(record);
        }
      }
    }, {
      key: '_getMany',
      value: function _getMany(type, params) {
        var _this = this;

        var typeStore = this._data[type];
        if (typeStore) {
          var keys = Object.keys(typeStore);
          if (keys.length) {
            return keys.map(function (key) {
              return _this._getByKey(type, key);
            });
          }
        }
      }
    }, {
      key: '_pushOne',
      value: function _pushOne(type, record) {
        record = this._applyTransforms(type, record, 'one');
        if (record) {
          var key = record[this._primaryKeyForType(type)];
          if (typeof key !== 'undefined') {
            key = key.toString();
            this._data[type][key] = record;
            return this._getByKey(type, key);
          }
        }
      }
    }, {
      key: '_pushMany',
      value: function _pushMany(type, records) {
        var _this2 = this;

        records = this._applyTransforms(type, records, 'many');
        if (records && records.length) {
          return records.map(function (record) {
            return _this2._pushOne(type, record);
          });
        }
      }
    }, {
      key: '_findByKey',
      value: function _findByKey(type, key) {
        var cached = this._getByKey(type, key);
        if (cached) {
          return Promise.resolve(cached);
        }
        return this._fetchByKey(type, key);
      }
    }, {
      key: '_findMany',
      value: function _findMany(type, params) {
        var cached = this._getMany(type, params);
        if (cached && cached.length) {
          return Promise.resolve(cached);
        }
        return this._fetchMany(type, params);
      }
    }, {
      key: '_fetchByKey',
      value: function _fetchByKey(type, key) {
        var _this3 = this;

        return this.adapter.get(type, key).then(function (data) {
          return _this3._pushOne(type, data);
        });
      }
    }, {
      key: '_fetchMany',
      value: function _fetchMany(type, params) {
        var _this4 = this;

        return this.adapter.get(type, params).then(function (data) {
          return _this4._pushMany(type, data);
        });
      }
    }, {
      key: '_applyTransforms',
      value: function _applyTransforms(type, record, requestType) {
        var _this5 = this;

        if (this._transforms.length) {
          (function () {
            var typeMeta = _this5.lookup(type);
            _this5._transforms.forEach(function (fnc) {
              record = fnc(record, typeMeta, requestType);
            });
          })();
        }
        return record;
      }
    }]);
    return Store;
  }();

  exports.Store = Store;
  exports.RESTAdapter = RESTAdapter;
  exports.FetchAdapter = FetchAdapter;
  exports.XHRAdapter = XHRAdapter;
  exports.HTTPAdapter = HTTPAdapter;

}));